@@ -1,10 +1,11 @@
 /* 
    Mount helper utility for Linux CIFS VFS (virtual filesystem) client
    Copyright (C) 2003,2008 Steve French  (sfrench@us.ibm.com)
+   Copyright (C) 2008 Jeremy Allison (jra@samba.org)
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
@@ -13,8 +14,7 @@
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
 #ifndef _GNU_SOURCE
 #define _GNU_SOURCE
@@ -41,7 +41,7 @@
 #include <limits.h>
 
 #define MOUNT_CIFS_VERSION_MAJOR "1"
-#define MOUNT_CIFS_VERSION_MINOR "10"
+#define MOUNT_CIFS_VERSION_MINOR "11"
 
 #ifndef MOUNT_CIFS_VENDOR_SUFFIX
  #ifdef _SAMBA_BUILD_
@@ -64,8 +64,17 @@
 #define MS_BIND 4096
 #endif
 
+#define MAX_UNC_LEN 1024
+
 #define CONST_DISCARD(type, ptr)      ((type) ((void *) (ptr)))
 
+#ifndef SAFE_FREE
+#define SAFE_FREE(x) do { if ((x) != NULL) {free(x); x=NULL;} } while(0)
+#endif
+
+#define MOUNT_PASSWD_SIZE 64
+#define DOMAIN_SIZE 64
+
 const char *thisprogram;
 int verboseflag = 0;
 static int got_password = 0;
@@ -75,7 +84,6 @@ static int got_ip = 0;
 static int got_unc = 0;
 static int got_uid = 0;
 static int got_gid = 0;
-static int free_share_name = 0;
 static char * user_name = NULL;
 static char * mountpassword = NULL;
 char * domain_name = NULL;
@@ -86,7 +94,7 @@ char * prefixpath = NULL;
 
 /* like strncpy but does not 0 fill the buffer and always null
  *    terminates. bufsize is the size of the destination buffer */
-size_t strlcpy(char *d, const char *s, size_t bufsize)
+static size_t strlcpy(char *d, const char *s, size_t bufsize)
 {
 	size_t len = strlen(s);
 	size_t ret = len;
@@ -100,13 +108,16 @@ size_t strlcpy(char *d, const char *s, size_t bufsize)
 /* like strncat but does not 0 fill the buffer and always null
  *    terminates. bufsize is the length of the buffer, which should
  *       be one more than the maximum resulting string length */
-size_t strlcat(char *d, const char *s, size_t bufsize)
+static size_t strlcat(char *d, const char *s, size_t bufsize)
 {
 	size_t len1 = strlen(d);
 	size_t len2 = strlen(s);
 	size_t ret = len1 + len2;
 
 	if (len1+len2 >= bufsize) {
+		if (bufsize < (len1+1)) {
+			return ret;
+		}
 		len2 = bufsize - (len1+1);
 	}
 	if (len2 > 0) {
@@ -137,24 +148,22 @@ static void mount_cifs_usage(void)
 	printf("\nLess commonly used options:");
 	printf("\n\tcredentials=<filename>,guest,perm,noperm,setuids,nosetuids,rw,ro,");
 	printf("\n\tsep=<char>,iocharset=<codepage>,suid,nosuid,exec,noexec,serverino,");
-	printf("\n\tdirectio,mapchars,nomapchars,nolock,servernetbiosname=<SRV_RFC1001NAME>");
+	printf("\n\tmapchars,nomapchars,nolock,servernetbiosname=<SRV_RFC1001NAME>");
+	printf("\n\tdirectio,nounix,cifsacl,sec=<authentication mechanism>,sign");
 	printf("\n\nOptions not needed for servers supporting CIFS Unix extensions");
 	printf("\n\t(e.g. unneeded for mounts to most Samba versions):");
 	printf("\n\tuid=<uid>,gid=<gid>,dir_mode=<mode>,file_mode=<mode>,sfu");
 	printf("\n\nRarely used options:");
 	printf("\n\tport=<tcpport>,rsize=<size>,wsize=<size>,unc=<unc_name>,ip=<ip_address>,");
 	printf("\n\tdev,nodev,nouser_xattr,netbiosname=<OUR_RFC1001NAME>,hard,soft,intr,");
-	printf("\n\tnointr,ignorecase,noposixpaths,noacl");
+	printf("\n\tnointr,ignorecase,noposixpaths,noacl,prefixpath=<path>,nobrl");
+	printf("\n\tin6_addr");
 	printf("\n\nOptions are described in more detail in the manual page");
 	printf("\n\tman 8 mount.cifs\n");
 	printf("\nTo display the version number of the mount helper:");
 	printf("\n\t%s -V\n",thisprogram);
 
-	if(mountpassword) {
-		memset(mountpassword,0,64);
-		free(mountpassword);
-		mountpassword = NULL;
-	}
+	SAFE_FREE(mountpassword);
 	exit(1);
 }
 
@@ -187,7 +196,7 @@ static int open_cred_file(char * file_name)
 	line_buf = (char *)malloc(4096);
 	if(line_buf == NULL) {
 		fclose(fs);
-		return -ENOMEM;
+		return ENOMEM;
 	}
 
 	while(fgets(line_buf,4096,fs)) {
@@ -207,22 +216,20 @@ static int open_cred_file(char * file_name)
 				for(length = 0;length<4087;length++) {
 					if ((temp_val[length] == '\n')
 					    || (temp_val[length] == '\0')) {
+						temp_val[length] = '\0';
 						break;
 					}
 				}
 				if(length > 4086) {
 					printf("mount.cifs failed due to malformed username in credentials file");
 					memset(line_buf,0,4096);
-					if(mountpassword) {
-						memset(mountpassword,0,64);
-					}
 					exit(1);
 				} else {
 					got_user = 1;
 					user_name = (char *)calloc(1 + length,1);
 					/* BB adding free of user_name string before exit,
 						not really necessary but would be cleaner */
-					strncpy(user_name,temp_val, length);
+					strlcpy(user_name,temp_val, length+1);
 				}
 			}
 		} else if (strncasecmp("password",line_buf+i,8) == 0) {
@@ -230,26 +237,24 @@ static int open_cred_file(char * file_name)
 			if(temp_val) {
 				/* go past equals sign */
 				temp_val++;
-				for(length = 0;length<65;length++) {
+				for(length = 0;length<MOUNT_PASSWD_SIZE+1;length++) {
 					if ((temp_val[length] == '\n')
 					    || (temp_val[length] == '\0')) {
+						temp_val[length] = '\0';
 						break;
 					}
 				}
-				if(length > 64) {
+				if(length > MOUNT_PASSWD_SIZE) {
 					printf("mount.cifs failed: password in credentials file too long\n");
 					memset(line_buf,0, 4096);
-					if(mountpassword) {
-						memset(mountpassword,0,64);
-					}
 					exit(1);
 				} else {
 					if(mountpassword == NULL) {
-						mountpassword = (char *)calloc(65,1);
+						mountpassword = (char *)calloc(MOUNT_PASSWD_SIZE+1,1);
 					} else
-						memset(mountpassword,0,64);
+						memset(mountpassword,0,MOUNT_PASSWD_SIZE);
 					if(mountpassword) {
-						strncpy(mountpassword,temp_val,length);
+						strlcpy(mountpassword,temp_val,MOUNT_PASSWD_SIZE+1);
 						got_password = 1;
 					}
 				}
@@ -261,25 +266,23 @@ static int open_cred_file(char * file_name)
                                 temp_val++;
 				if(verboseflag)
 					printf("\nDomain %s\n",temp_val);
-                                for(length = 0;length<65;length++) {
+                                for(length = 0;length<DOMAIN_SIZE+1;length++) {
 					if ((temp_val[length] == '\n')
 					    || (temp_val[length] == '\0')) {
+						temp_val[length] = '\0';
 						break;
 					}
                                 }
-                                if(length > 64) {
+                                if(length > DOMAIN_SIZE) {
                                         printf("mount.cifs failed: domain in credentials file too long\n");
-                                        if(mountpassword) {
-                                                memset(mountpassword,0,64);
-                                        }
                                         exit(1);
                                 } else {
                                         if(domain_name == NULL) {
-                                                domain_name = (char *)calloc(65,1);
+                                                domain_name = (char *)calloc(DOMAIN_SIZE+1,1);
                                         } else
-                                                memset(domain_name,0,64);
+                                                memset(domain_name,0,DOMAIN_SIZE);
                                         if(domain_name) {
-                                                strncpy(domain_name,temp_val,length);
+                                                strlcpy(domain_name,temp_val,DOMAIN_SIZE+1);
                                                 got_domain = 1;
                                         }
                                 }
@@ -288,11 +291,7 @@ static int open_cred_file(char * file_name)
 
 	}
 	fclose(fs);
-	if(line_buf) {
-		memset(line_buf,0,4096);
-		free(line_buf);
-		line_buf = NULL;
-	}
+	SAFE_FREE(line_buf);
 	return 0;
 }
 
@@ -303,9 +302,9 @@ static int get_password_from_file(int file_descript, char * filename)
 	char c;
 
 	if(mountpassword == NULL)
-		mountpassword = (char *)calloc(65,1);
+		mountpassword = (char *)calloc(MOUNT_PASSWD_SIZE+1,1);
 	else 
-		memset(mountpassword, 0, 64);
+		memset(mountpassword, 0, MOUNT_PASSWD_SIZE);
 
 	if (mountpassword == NULL) {
 		printf("malloc failed\n");
@@ -322,11 +321,10 @@ static int get_password_from_file(int file_descript, char * filename)
 	}
 	/* else file already open and fd provided */
 
-	for(i=0;i<64;i++) {
+	for(i=0;i<MOUNT_PASSWD_SIZE;i++) {
 		rc = read(file_descript,&c,1);
 		if(rc < 0) {
 			printf("mount.cifs failed. Error %s reading password file\n",strerror(errno));
-			memset(mountpassword,0,64);
 			if(filename != NULL)
 				close(file_descript);
 			exit(1);
@@ -338,13 +336,15 @@ static int get_password_from_file(int file_descript, char * filename)
 			break;
 		} else /* read valid character */ {
 			if((c == 0) || (c == '\n')) {
+				mountpassword[i] = '\0';
 				break;
 			} else 
 				mountpassword[i] = c;
 		}
 	}
-	if((i == 64) && (verboseflag)) {
-		printf("\nWarning: password longer than 64 characters specified in cifs password file");
+	if((i == MOUNT_PASSWD_SIZE) && (verboseflag)) {
+		printf("\nWarning: password longer than %d characters specified in cifs password file",
+			MOUNT_PASSWD_SIZE);
 	}
 	got_password = 1;
 	if(filename != NULL) {
@@ -431,13 +431,13 @@ static int parse_options(char ** optionsp, int * filesys_flags)
 					if(percent_char) {
 						*percent_char = ',';
 						if(mountpassword == NULL)
-							mountpassword = (char *)calloc(65,1);
+							mountpassword = (char *)calloc(MOUNT_PASSWD_SIZE+1,1);
 						if(mountpassword) {
 							if(got_password)
 								printf("\nmount.cifs warning - password specified twice\n");
 							got_password = 1;
 							percent_char++;
-							strncpy(mountpassword, percent_char,64);
+							strlcpy(mountpassword, percent_char,MOUNT_PASSWD_SIZE+1);
 						/*  remove password from username */
 							while(*percent_char != 0) {
 								*percent_char = ',';
@@ -523,7 +523,7 @@ static int parse_options(char ** optionsp, int * filesys_flags)
 				printf("CIFS: invalid domain name\n");
 				return 1;	/* needs_arg; */
 			}
-			if (strnlen(value, 65) < 65) {
+			if (strnlen(value, DOMAIN_SIZE+1) < DOMAIN_SIZE+1) {
 				got_domain = 1;
 			} else {
 				printf("domain name too long\n");
@@ -533,7 +533,8 @@ static int parse_options(char ** optionsp, int * filesys_flags)
 			if (value && *value) {
 				rc = open_cred_file(value);
 				if(rc) {
-					printf("error %d opening credential file %s\n",rc, value);
+					printf("error %d (%s) opening credential file %s\n",
+						rc, strerror(rc), value);
 					return 1;
 				}
 			} else {
@@ -711,7 +712,7 @@ nocopy:
 		out_len = strlen(out);
 	}
 
-	free(*optionsp);
+	SAFE_FREE(*optionsp);
 	*optionsp = out;
 	return 0;
 }
@@ -739,7 +740,7 @@ static void check_for_comma(char ** ppasswrd)
 
 	if(number_of_commas == 0)
 		return;
-	if(number_of_commas > 64) {
+	if(number_of_commas > MOUNT_PASSWD_SIZE) {
 		/* would otherwise overflow the mount options buffer */
 		printf("\nInvalid password. Password contains too many commas.\n");
 		return;
@@ -758,7 +759,7 @@ static void check_for_comma(char ** ppasswrd)
 	}
 	new_pass_buf[len+number_of_commas] = 0;
 
-	free(*ppasswrd);
+	SAFE_FREE(*ppasswrd);
 	*ppasswrd = new_pass_buf;
 	
 	return;
@@ -838,17 +839,31 @@ static char * check_for_domain(char **ppuser)
 	return domainnm;
 }
 
+/* replace all occurances of "from" in a string with "to" */
+static void replace_char(char *string, char from, char to, int maxlen)
+{
+	char *lastchar = string + maxlen;
+	while (string) {
+		string = strchr(string, from);
+		if (string) {
+			*string = to;
+			if (string >= lastchar)
+				return;
+		}
+	}
+}
+
 /* Note that caller frees the returned buffer if necessary */
 static char * parse_server(char ** punc_name)
 {
 	char * unc_name = *punc_name;
-	int length = strnlen(unc_name,1024);
+	int length = strnlen(unc_name, MAX_UNC_LEN);
 	char * share;
 	char * ipaddress_string = NULL;
 	struct hostent * host_entry = NULL;
 	struct in_addr server_ipaddr;
 
-	if(length > 1023) {
+	if(length > (MAX_UNC_LEN - 1)) {
 		printf("mount error: UNC name too long");
 		return NULL;
 	}
@@ -867,7 +882,6 @@ static char * parse_server(char ** punc_name)
 			/* check for nfs syntax ie server:share */
 			share = strchr(unc_name,':');
 			if(share) {
-				free_share_name = 1;
 				*punc_name = (char *)malloc(length+3);
 				if(*punc_name == NULL) {
 					/* put the original string back  if 
@@ -875,9 +889,9 @@ static char * parse_server(char ** punc_name)
 					*punc_name = unc_name;
 					return NULL;
 				}
-					
 				*share = '/';
-				strncpy((*punc_name)+2,unc_name,length);
+				strlcpy((*punc_name)+2,unc_name,length+1);
+				SAFE_FREE(unc_name);
 				unc_name = *punc_name;
 				unc_name[length+2] = 0;
 				goto continue_unc_parsing;
@@ -891,15 +905,18 @@ continue_unc_parsing:
 			unc_name[0] = '/';
 			unc_name[1] = '/';
 			unc_name += 2;
-			if ((share = strchr(unc_name, '/')) || 
-				(share = strchr(unc_name,'\\'))) {
+
+			/* allow for either delimiter between host and sharename */
+			if ((share = strpbrk(unc_name, "/\\"))) {
 				*share = 0;  /* temporarily terminate the string */
 				share += 1;
 				if(got_ip == 0) {
 					host_entry = gethostbyname(unc_name);
 				}
-				*(share - 1) = '/'; /* put the slash back */
-				if ((prefixpath = strchr(share, '/'))) {
+				*(share - 1) = '/'; /* put delimiter back */
+
+				/* we don't convert the prefixpath delimiters since '\\' is a valid char in posix paths */
+				if ((prefixpath = strpbrk(share, "/\\"))) {
 					*prefixpath = 0;  /* permanently terminate the string */
 					if (!strlen(++prefixpath))
 						prefixpath = NULL; /* this needs to be done explicitly */
@@ -964,6 +981,25 @@ static struct option longopts[] = {
 	{ NULL, 0, NULL, 0 }
 };
 
+/* convert a string to uppercase. return false if the string
+ * wasn't ASCII or was a NULL ptr */
+static int
+uppercase_string(char *string)
+{
+	if (!string)
+		return 0;
+
+	while (*string) {
+		/* check for unicode */
+		if ((unsigned char) string[0] & 0x80)
+			return 0;
+		*string = toupper((unsigned char) *string);
+		string++;
+	}
+
+	return 1;
+}
+
 int main(int argc, char ** argv)
 {
 	int c;
@@ -976,6 +1012,7 @@ int main(int argc, char ** argv)
 	char * options = NULL;
 	char * resolved_path = NULL;
 	char * temp;
+	char * dev_name;
 	int rc;
 	int rsize = 0;
 	int wsize = 0;
@@ -1012,8 +1049,16 @@ int main(int argc, char ** argv)
 	printf(" node: %s machine: %s sysname %s domain %s\n", sysinfo.nodename,sysinfo.machine,sysinfo.sysname,sysinfo.domainname);
 #endif */
 	if(argc > 2) {
-		share_name = argv[1];
+		dev_name = argv[1];
+		share_name = strndup(argv[1], MAX_UNC_LEN);
+		if (share_name == NULL) {
+			fprintf(stderr, "%s: %s", argv[0], strerror(ENOMEM));
+			exit(1);
+		}
 		mountpoint = argv[2];
+	} else {
+		mount_cifs_usage();
+		exit(1);
 	}
 
 	/* add sharename in opts string as unc= parm */
@@ -1072,9 +1117,6 @@ int main(int argc, char ** argv)
 			MOUNT_CIFS_VERSION_MAJOR,
 			MOUNT_CIFS_VERSION_MINOR,
 			MOUNT_CIFS_VENDOR_SUFFIX);
-			if(mountpassword) {
-				memset(mountpassword,0,64);
-			}
 			exit (0);
 		case 'w':
 			flags &= ~MS_RDONLY;
@@ -1135,10 +1177,10 @@ int main(int argc, char ** argv)
 			break;
 		case 'p':
 			if(mountpassword == NULL)
-				mountpassword = (char *)calloc(65,1);
+				mountpassword = (char *)calloc(MOUNT_PASSWD_SIZE+1,1);
 			if(mountpassword) {
 				got_password = 1;
-				strncpy(mountpassword,optarg,64);
+				strlcpy(mountpassword,optarg,MOUNT_PASSWD_SIZE+1);
 			}
 			break;
 		case 'S':
@@ -1153,16 +1195,16 @@ int main(int argc, char ** argv)
 		}
 	}
 
-	if((argc < 3) || (share_name == NULL) || (mountpoint == NULL)) {
+	if((argc < 3) || (dev_name == NULL) || (mountpoint == NULL)) {
 		mount_cifs_usage();
 		exit(1);
 	}
 
 	if (getenv("PASSWD")) {
 		if(mountpassword == NULL)
-			mountpassword = (char *)calloc(65,1);
+			mountpassword = (char *)calloc(MOUNT_PASSWD_SIZE+1,1);
 		if(mountpassword) {
-			strncpy(mountpassword,getenv("PASSWD"),64);
+			strlcpy(mountpassword,getenv("PASSWD"),MOUNT_PASSWD_SIZE+1);
 			got_password = 1;
 		}
 	} else if (getenv("PASSWD_FD")) {
@@ -1228,16 +1270,14 @@ int main(int argc, char ** argv)
 	}
        
 	if(got_password == 0) {
-		char *tmp_pass;
-		tmp_pass = getpass("Password: "); /* BB obsolete sys call but
-						     no good replacement yet */
-		mountpassword = (char *)calloc(65,1);
+		char *tmp_pass = getpass("Password: "); /* BB obsolete sys call but
+							   no good replacement yet. */
+		mountpassword = (char *)calloc(MOUNT_PASSWD_SIZE+1,1);
 		if (!tmp_pass || !mountpassword) {
-			printf("Password not entered, exiting.\n");
+			printf("Password not entered, exiting\n");
 			return -1;
 		}
-		strncpy(mountpassword, tmp_pass, 64);
-						 
+		strlcpy(mountpassword, tmp_pass, MOUNT_PASSWD_SIZE+1);
 		got_password = 1;
 	}
 	/* FIXME launch daemon (handles dfs name resolution and credential change) 
@@ -1261,11 +1301,8 @@ mount_retry:
 		optlen += strlen(ipaddr) + 4;
 	if(mountpassword)
 		optlen += strlen(mountpassword) + 6;
-	if(options) {
-		free(options);
-		options = NULL;
-	}
-	options_size = optlen + 10 + 64;
+	SAFE_FREE(options);
+	options_size = optlen + 10 + DOMAIN_SIZE;
 	options = (char *)malloc(options_size /* space for commas in password */ + 8 /* space for domain=  , domain name itself was counted as part of the length username string above */);
 
 	if(options == NULL) {
@@ -1322,10 +1359,12 @@ mount_retry:
 	}
 	if(verboseflag)
 		printf("\nmount.cifs kernel mount options %s \n",options);
-	if(mount(share_name, mountpoint, "cifs", flags, options)) {
-	/* remember to kill daemon on error */
-		char * tmp;
 
+	/* convert all '\\' to '/' in share portion so that /proc/mounts looks pretty */
+	replace_char(dev_name, '\\', '/', strlen(share_name));
+
+	if(mount(dev_name, mountpoint, "cifs", flags, options)) {
+	/* remember to kill daemon on error */
 		switch (errno) {
 		case 0:
 			printf("mount failed but no error number set\n");
@@ -1336,12 +1375,9 @@ mount_retry:
 		case ENXIO:
 			if(retry == 0) {
 				retry = 1;
-				tmp = share_name;
-				while (*tmp && !(((unsigned char)tmp[0]) & 0x80)) {
-					*tmp = toupper((unsigned char)*tmp);
-		        		tmp++;
-				}
-				if(!*tmp) {
+				if (uppercase_string(dev_name) &&
+				    uppercase_string(share_name) &&
+				    uppercase_string(prefixpath)) {
 					printf("retrying with upper case share name\n");
 					goto mount_retry;
 				}
@@ -1355,9 +1391,9 @@ mount_retry:
 	} else {
 		pmntfile = setmntent(MOUNTED, "a+");
 		if(pmntfile) {
-			mountent.mnt_fsname = share_name;
-			mountent.mnt_dir = mountpoint; 
-			mountent.mnt_type = CONST_DISCARD(char *,"cifs"); 
+			mountent.mnt_fsname = dev_name;
+			mountent.mnt_dir = mountpoint;
+			mountent.mnt_type = CONST_DISCARD(char *,"cifs");
 			mountent.mnt_opts = (char *)malloc(220);
 			if(mountent.mnt_opts) {
 				char * mount_user = getusername();
@@ -1388,10 +1424,7 @@ mount_retry:
 			mountent.mnt_passno = 0;
 			rc = addmntent(pmntfile,&mountent);
 			endmntent(pmntfile);
-			if(mountent.mnt_opts) {
-				free(mountent.mnt_opts);
-				mountent.mnt_opts = NULL;
-			}
+			SAFE_FREE(mountent.mnt_opts);
 		} else {
 		    printf("could not update mount table\n");
 		}
@@ -1401,29 +1434,12 @@ mount_exit:
 	if(mountpassword) {
 		int len = strlen(mountpassword);
 		memset(mountpassword,0,len);
-		free(mountpassword);
-		mountpassword = NULL;
+		SAFE_FREE(mountpassword);
 	}
 
-	if(options) {
-		memset(options,0,optlen);
-		free(options);
-		options = NULL;
-	}
-
-	if(orgoptions) {
-		memset(orgoptions,0,orgoptlen);
-		free(orgoptions);
-		orgoptions = NULL;
-	}
-	if(resolved_path) {
-		free(resolved_path);
-		resolved_path = NULL;
-	}
-
-	if(free_share_name) {
-		free(share_name);
-		share_name = NULL;
-		}
+	SAFE_FREE(options);
+	SAFE_FREE(orgoptions);
+	SAFE_FREE(resolved_path);
+	SAFE_FREE(share_name);
 	return rc;
 }
