@@ -8,7 +8,7 @@
  *  
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
+ *  the Free Software Foundation; either version 3 of the License, or
  *  (at your option) any later version.
  *  
  *  This program is distributed in the hope that it will be useful,
@@ -17,13 +17,47 @@
  *  GNU General Public License for more details.
  *  
  *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *  along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "includes.h"
 #include "groupdb/mapping.h"
 
+static const struct mapping_backend *backend;
+
+/*
+  initialise a group mapping backend
+ */
+static bool init_group_mapping(void)
+{
+	const char *backend_string;
+
+	if (backend != NULL) {
+		/* already initialised */
+		return True;
+	}
+
+	/*
+	 * default to using the ldb backend. This parameter should
+	 * disappear in future versions of Samba3.
+	 *
+	 * But it's needed for cluster setups, because it's
+	 * not yet possible to distribute a ldb inside a cluster.
+	 */
+	backend_string = lp_parm_const_string(-1, "groupdb", "backend", "ldb");
+
+	if (strcmp(backend_string, "ldb") == 0) {
+		backend = groupdb_ldb_init();
+	} else if (strcmp(backend_string, "tdb") == 0) {
+		backend = groupdb_tdb_init();
+	} else {
+		DEBUG(0,("Unknown groupdb backend '%s'\n", backend_string));
+		smb_panic("Unknown groupdb backend");
+	}
+
+	return backend != NULL;
+}
+
 /****************************************************************************
 initialise first time the mapping list
 ****************************************************************************/
@@ -49,70 +83,6 @@ NTSTATUS add_initial_entry(gid_t gid, const char *sid, enum lsa_SidType sid_name
 	return pdb_add_group_mapping_entry(&map);
 }
 
-/****************************************************************************
- Map a unix group to a newly created mapping
-****************************************************************************/
-NTSTATUS map_unix_group(const struct group *grp, GROUP_MAP *pmap)
-{
-	NTSTATUS status;
-	GROUP_MAP map;
-	const char *grpname, *dom, *name;
-	uint32 rid;
-
-	if (pdb_getgrgid(&map, grp->gr_gid)) {
-		return NT_STATUS_GROUP_EXISTS;
-	}
-
-	map.gid = grp->gr_gid;
-	grpname = grp->gr_name;
-
-	if (lookup_name(tmp_talloc_ctx(), grpname, LOOKUP_NAME_ISOLATED,
-			&dom, &name, NULL, NULL)) {
-
-		const char *tmp = talloc_asprintf(
-			tmp_talloc_ctx(), "Unix Group %s", grp->gr_name);
-
-		DEBUG(5, ("%s exists as %s\\%s, retrying as \"%s\"\n",
-			  grpname, dom, name, tmp));
-		grpname = tmp;
-	}
-
-	if (lookup_name(tmp_talloc_ctx(), grpname, LOOKUP_NAME_ISOLATED,
-			NULL, NULL, NULL, NULL)) {
-		DEBUG(3, ("\"%s\" exists, can't map it\n", grp->gr_name));
-		return NT_STATUS_GROUP_EXISTS;
-	}
-
-	fstrcpy(map.nt_name, grpname);
-
-	if (pdb_rid_algorithm()) {
-		rid = algorithmic_pdb_gid_to_group_rid( grp->gr_gid );
-	} else {
-		if (!pdb_new_rid(&rid)) {
-			DEBUG(3, ("Could not get a new RID for %s\n",
-				  grp->gr_name));
-			return NT_STATUS_ACCESS_DENIED;
-		}
-	}
-
-	sid_compose(&map.sid, get_global_sam_sid(), rid);
-	map.sid_name_use = SID_NAME_DOM_GRP;
-	fstrcpy(map.comment, talloc_asprintf(tmp_talloc_ctx(), "Unix Group %s",
-					     grp->gr_name));
-
-	status = pdb_add_group_mapping_entry(&map);
-	if (NT_STATUS_IS_OK(status)) {
-		*pmap = map;
-	}
-	return status;
-}
-
-
-
-
-
-
-
 static NTSTATUS alias_memberships(const DOM_SID *members, size_t num_members,
 				  DOM_SID **sids, size_t *num)
 {
@@ -122,7 +92,7 @@ static NTSTATUS alias_memberships(const DOM_SID *members, size_t num_members,
 	*sids = NULL;
 
 	for (i=0; i<num_members; i++) {
-		NTSTATUS status = one_alias_membership(&members[i], sids, num);
+		NTSTATUS status = backend->one_alias_membership(&members[i], sids, num);
 		if (!NT_STATUS_IS_OK(status))
 			return status;
 	}
@@ -149,10 +119,10 @@ struct aliasmem_closure {
 
 /* get a domain group from it's SID */
 
-BOOL get_domain_group_from_sid(DOM_SID sid, GROUP_MAP *map)
+bool get_domain_group_from_sid(DOM_SID sid, GROUP_MAP *map)
 {
 	struct group *grp;
-	BOOL ret;
+	bool ret;
 	
 	if(!init_group_mapping()) {
 		DEBUG(0,("failed to initialize group mapping\n"));
@@ -179,6 +149,7 @@ BOOL get_domain_group_from_sid(DOM_SID sid, GROUP_MAP *map)
 			fstrcpy( map->comment, "Ordinary Users" );
 			sid_copy( &map->sid, &sid );
 			map->sid_name_use = SID_NAME_DOM_GRP;
+			map->gid = (gid_t)-1;
 			
 			return True;
 		}
@@ -218,17 +189,28 @@ BOOL get_domain_group_from_sid(DOM_SID sid, GROUP_MAP *map)
 
 int smb_create_group(const char *unix_group, gid_t *new_gid)
 {
-	pstring add_script;
+	char *add_script = NULL;
 	int 	ret = -1;
 	int 	fd = 0;
-	
+
 	*new_gid = 0;
 
 	/* defer to scripts */
-	
+
 	if ( *lp_addgroup_script() ) {
-		pstrcpy(add_script, lp_addgroup_script());
-		pstring_sub(add_script, "%g", unix_group);
+		TALLOC_CTX *ctx = talloc_tos();
+
+		add_script = talloc_strdup(ctx,
+					lp_addgroup_script());
+		if (!add_script) {
+			return -1;
+		}
+		add_script = talloc_string_sub(ctx,
+				add_script, "%g", unix_group);
+		if (!add_script) {
+			return -1;
+		}
+
 		ret = smbrun(add_script, &fd);
 		DEBUG(ret ? 0 : 3,("smb_create_group: Running the command `%s' gave %d\n",add_script,ret));
 		if (ret == 0) {
@@ -244,7 +226,7 @@ int smb_create_group(const char *unix_group, gid_t *new_gid)
 			if (read(fd, output, sizeof(output)) > 0) {
 				*new_gid = (gid_t)strtoul(output, NULL, 10);
 			}
-			
+
 			close(fd);
 		}
 
@@ -256,8 +238,8 @@ int smb_create_group(const char *unix_group, gid_t *new_gid)
 		if (grp != NULL)
 			*new_gid = grp->gr_gid;
 	}
-			
-	return ret;	
+
+	return ret;
 }
 
 /****************************************************************************
@@ -266,14 +248,24 @@ int smb_create_group(const char *unix_group, gid_t *new_gid)
 
 int smb_delete_group(const char *unix_group)
 {
-	pstring del_script;
-	int ret;
+	char *del_script = NULL;
+	int ret = -1;
 
 	/* defer to scripts */
-	
+
 	if ( *lp_delgroup_script() ) {
-		pstrcpy(del_script, lp_delgroup_script());
-		pstring_sub(del_script, "%g", unix_group);
+		TALLOC_CTX *ctx = talloc_tos();
+
+		del_script = talloc_strdup(ctx,
+				lp_delgroup_script());
+		if (!del_script) {
+			return -1;
+		}
+		del_script = talloc_string_sub(ctx,
+				del_script, "%g", unix_group);
+		if (!del_script) {
+			return -1;
+		}
 		ret = smbrun(del_script,NULL);
 		DEBUG(ret ? 0 : 3,("smb_delete_group: Running the command `%s' gave %d\n",del_script,ret));
 		if (ret == 0) {
@@ -281,24 +273,39 @@ int smb_delete_group(const char *unix_group)
 		}
 		return ret;
 	}
-		
+
 	return -1;
 }
 
 /****************************************************************************
  Set a user's primary UNIX group.
 ****************************************************************************/
+
 int smb_set_primary_group(const char *unix_group, const char* unix_user)
 {
-	pstring add_script;
-	int ret;
+	char *add_script = NULL;
+	int ret = -1;
 
 	/* defer to scripts */
-	
+
 	if ( *lp_setprimarygroup_script() ) {
-		pstrcpy(add_script, lp_setprimarygroup_script());
-		all_string_sub(add_script, "%g", unix_group, sizeof(add_script));
-		all_string_sub(add_script, "%u", unix_user, sizeof(add_script));
+		TALLOC_CTX *ctx = talloc_tos();
+
+		add_script = talloc_strdup(ctx,
+				lp_setprimarygroup_script());
+		if (!add_script) {
+			return -1;
+		}
+		add_script = talloc_all_string_sub(ctx,
+				add_script, "%g", unix_group);
+		if (!add_script) {
+			return -1;
+		}
+		add_script = talloc_string_sub(ctx,
+				add_script, "%u", unix_user);
+		if (!add_script) {
+			return -1;
+		}
 		ret = smbrun(add_script,NULL);
 		flush_pwnam_cache();
 		DEBUG(ret ? 0 : 3,("smb_set_primary_group: "
@@ -318,15 +325,29 @@ int smb_set_primary_group(const char *unix_group, const char* unix_user)
 
 int smb_add_user_group(const char *unix_group, const char *unix_user)
 {
-	pstring add_script;
-	int ret;
+	char *add_script = NULL;
+	int ret = -1;
 
 	/* defer to scripts */
-	
+
 	if ( *lp_addusertogroup_script() ) {
-		pstrcpy(add_script, lp_addusertogroup_script());
-		pstring_sub(add_script, "%g", unix_group);
-		pstring_sub(add_script, "%u", unix_user);
+		TALLOC_CTX *ctx = talloc_tos();
+
+		add_script = talloc_strdup(ctx,
+				lp_addusertogroup_script());
+		if (!add_script) {
+			return -1;
+		}
+		add_script = talloc_string_sub(ctx,
+				add_script, "%g", unix_group);
+		if (!add_script) {
+			return -1;
+		}
+		add_script = talloc_string_sub(ctx,
+				add_script, "%u", unix_user);
+		if (!add_script) {
+			return -1;
+		}
 		ret = smbrun(add_script,NULL);
 		DEBUG(ret ? 0 : 3,("smb_add_user_group: Running the command `%s' gave %d\n",add_script,ret));
 		if (ret == 0) {
@@ -334,7 +355,7 @@ int smb_add_user_group(const char *unix_group, const char *unix_user)
 		}
 		return ret;
 	}
-	
+
 	return -1;
 }
 
@@ -344,15 +365,29 @@ int smb_add_user_group(const char *unix_group, const char *unix_user)
 
 int smb_delete_user_group(const char *unix_group, const char *unix_user)
 {
-	pstring del_script;
-	int ret;
+	char *del_script = NULL;
+	int ret = -1;
 
 	/* defer to scripts */
-	
+
 	if ( *lp_deluserfromgroup_script() ) {
-		pstrcpy(del_script, lp_deluserfromgroup_script());
-		pstring_sub(del_script, "%g", unix_group);
-		pstring_sub(del_script, "%u", unix_user);
+		TALLOC_CTX *ctx = talloc_tos();
+
+		del_script = talloc_strdup(ctx,
+				lp_deluserfromgroup_script());
+		if (!del_script) {
+			return -1;
+		}
+		del_script = talloc_string_sub(ctx,
+				del_script, "%g", unix_group);
+		if (!del_script) {
+			return -1;
+		}
+		del_script = talloc_string_sub(ctx,
+				del_script, "%u", unix_user);
+		if (!del_script) {
+			return -1;
+		}
 		ret = smbrun(del_script,NULL);
 		DEBUG(ret ? 0 : 3,("smb_delete_user_group: Running the command `%s' gave %d\n",del_script,ret));
 		if (ret == 0) {
@@ -360,7 +395,7 @@ int smb_delete_user_group(const char *unix_group, const char *unix_user)
 		}
 		return ret;
 	}
-	
+
 	return -1;
 }
 
@@ -368,70 +403,82 @@ int smb_delete_user_group(const char *unix_group, const char *unix_user)
 NTSTATUS pdb_default_getgrsid(struct pdb_methods *methods, GROUP_MAP *map,
 				 DOM_SID sid)
 {
-	return get_group_map_from_sid(sid, map) ?
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->get_group_map_from_sid(sid, map) ?
 		NT_STATUS_OK : NT_STATUS_UNSUCCESSFUL;
 }
 
 NTSTATUS pdb_default_getgrgid(struct pdb_methods *methods, GROUP_MAP *map,
 				 gid_t gid)
 {
-	return get_group_map_from_gid(gid, map) ?
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->get_group_map_from_gid(gid, map) ?
 		NT_STATUS_OK : NT_STATUS_UNSUCCESSFUL;
 }
 
 NTSTATUS pdb_default_getgrnam(struct pdb_methods *methods, GROUP_MAP *map,
 				 const char *name)
 {
-	return get_group_map_from_ntname(name, map) ?
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->get_group_map_from_ntname(name, map) ?
 		NT_STATUS_OK : NT_STATUS_UNSUCCESSFUL;
 }
 
 NTSTATUS pdb_default_add_group_mapping_entry(struct pdb_methods *methods,
 						GROUP_MAP *map)
 {
-	return add_mapping_entry(map, TDB_INSERT) ?
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->add_mapping_entry(map, TDB_INSERT) ?
 		NT_STATUS_OK : NT_STATUS_UNSUCCESSFUL;
 }
 
 NTSTATUS pdb_default_update_group_mapping_entry(struct pdb_methods *methods,
 						   GROUP_MAP *map)
 {
-	return add_mapping_entry(map, TDB_REPLACE) ?
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->add_mapping_entry(map, TDB_REPLACE) ?
 		NT_STATUS_OK : NT_STATUS_UNSUCCESSFUL;
 }
 
 NTSTATUS pdb_default_delete_group_mapping_entry(struct pdb_methods *methods,
 						   DOM_SID sid)
 {
-	return group_map_remove(&sid) ?
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->group_map_remove(&sid) ?
 		NT_STATUS_OK : NT_STATUS_UNSUCCESSFUL;
 }
 
 NTSTATUS pdb_default_enum_group_mapping(struct pdb_methods *methods,
 					   const DOM_SID *sid, enum lsa_SidType sid_name_use,
 					   GROUP_MAP **pp_rmap, size_t *p_num_entries,
-					   BOOL unix_only)
+					   bool unix_only)
 {
-	return enum_group_mapping(sid, sid_name_use, pp_rmap, p_num_entries, unix_only) ?
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->enum_group_mapping(sid, sid_name_use, pp_rmap, p_num_entries, unix_only) ?
 		NT_STATUS_OK : NT_STATUS_UNSUCCESSFUL;
 }
 
-NTSTATUS pdb_default_find_alias(struct pdb_methods *methods,
-				const char *name, DOM_SID *sid)
-{
-	GROUP_MAP map;
-
-	if (!pdb_getgrnam(&map, name))
-		return NT_STATUS_NO_SUCH_ALIAS;
-
-	if ((map.sid_name_use != SID_NAME_WKN_GRP) &&
-	    (map.sid_name_use != SID_NAME_ALIAS))
-		return NT_STATUS_OBJECT_TYPE_MISMATCH;
-
-	sid_copy(sid, &map.sid);
-	return NT_STATUS_OK;
-}
-
 NTSTATUS pdb_default_create_alias(struct pdb_methods *methods,
 				  const char *name, uint32 *rid)
 {
@@ -439,7 +486,7 @@ NTSTATUS pdb_default_create_alias(struct pdb_methods *methods,
 	enum lsa_SidType type;
 	uint32 new_rid;
 	gid_t gid;
-	BOOL exists;
+	bool exists;
 	GROUP_MAP map;
 	TALLOC_CTX *mem_ctx;
 	NTSTATUS status;
@@ -451,7 +498,7 @@ NTSTATUS pdb_default_create_alias(struct pdb_methods *methods,
 		return NT_STATUS_NO_MEMORY;
 	}
 
-	exists = lookup_name(mem_ctx, name, LOOKUP_NAME_ISOLATED,
+	exists = lookup_name(mem_ctx, name, LOOKUP_NAME_LOCAL,
 			     NULL, NULL, &sid, &type);
 	TALLOC_FREE(mem_ctx);
 
@@ -512,7 +559,7 @@ NTSTATUS pdb_default_get_aliasinfo(struct pdb_methods *methods,
 	if ((map.sid_name_use != SID_NAME_ALIAS) &&
 	    (map.sid_name_use != SID_NAME_WKN_GRP)) {
 		DEBUG(2, ("%s is a %s, expected an alias\n",
-			  sid_string_static(sid),
+			  sid_string_dbg(sid),
 			  sid_type_lookup(map.sid_name_use)));
 		return NT_STATUS_NO_SUCH_ALIAS;
 	}
@@ -541,20 +588,32 @@ NTSTATUS pdb_default_set_aliasinfo(struct pdb_methods *methods,
 NTSTATUS pdb_default_add_aliasmem(struct pdb_methods *methods,
 				  const DOM_SID *alias, const DOM_SID *member)
 {
-	return add_aliasmem(alias, member);
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->add_aliasmem(alias, member);
 }
 
 NTSTATUS pdb_default_del_aliasmem(struct pdb_methods *methods,
 				  const DOM_SID *alias, const DOM_SID *member)
 {
-	return del_aliasmem(alias, member);
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->del_aliasmem(alias, member);
 }
 
 NTSTATUS pdb_default_enum_aliasmem(struct pdb_methods *methods,
 				   const DOM_SID *alias, DOM_SID **pp_members,
 				   size_t *p_num_members)
 {
-	return enum_aliasmem(alias, pp_members, p_num_members);
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+	return backend->enum_aliasmem(alias, pp_members, p_num_members);
 }
 
 NTSTATUS pdb_default_alias_memberships(struct pdb_methods *methods,
@@ -569,6 +628,11 @@ NTSTATUS pdb_default_alias_memberships(struct pdb_methods *methods,
 	size_t i, num_alias_sids;
 	NTSTATUS result;
 
+	if (!init_group_mapping()) {
+		DEBUG(0,("failed to initialize group mapping\n"));
+		return NT_STATUS_UNSUCCESSFUL;
+	}
+
 	alias_sids = NULL;
 	num_alias_sids = 0;
 
@@ -644,7 +708,7 @@ NTSTATUS pdb_nop_delete_group_mapping_entry(struct pdb_methods *methods,
 NTSTATUS pdb_nop_enum_group_mapping(struct pdb_methods *methods,
 					   enum lsa_SidType sid_name_use,
 					   GROUP_MAP **rmap, size_t *num_entries,
-					   BOOL unix_only)
+					   bool unix_only)
 {
 	return NT_STATUS_UNSUCCESSFUL;
 }
@@ -652,10 +716,10 @@ NTSTATUS pdb_nop_enum_group_mapping(struct pdb_methods *methods,
 /****************************************************************************
  These need to be redirected through pdb_interface.c
 ****************************************************************************/
-BOOL pdb_get_dom_grp_info(const DOM_SID *sid, struct acct_info *info)
+bool pdb_get_dom_grp_info(const DOM_SID *sid, struct acct_info *info)
 {
 	GROUP_MAP map;
-	BOOL res;
+	bool res;
 
 	become_root();
 	res = get_domain_group_from_sid(*sid, &map);
@@ -670,7 +734,7 @@ BOOL pdb_get_dom_grp_info(const DOM_SID *sid, struct acct_info *info)
 	return True;
 }
 
-BOOL pdb_set_dom_grp_info(const DOM_SID *sid, const struct acct_info *info)
+bool pdb_set_dom_grp_info(const DOM_SID *sid, const struct acct_info *info)
 {
 	GROUP_MAP map;
 
@@ -723,12 +787,12 @@ NTSTATUS pdb_create_builtin_alias(uint32 rid)
 		return NT_STATUS_ACCESS_DENIED;
 	}
 
-	DEBUG(10,("Creating alias %s with gid %d\n", name, gid));
+	DEBUG(10,("Creating alias %s with gid %d\n", groupname, gid));
 
 	map.gid = gid;
 	sid_copy(&map.sid, &sid);
 	map.sid_name_use = SID_NAME_ALIAS;
-	fstrcpy(map.nt_name, name);
+	fstrcpy(map.nt_name, groupname);
 	fstrcpy(map.comment, "");
 
 	status = pdb_add_group_mapping_entry(&map);
