@@ -7,19 +7,17 @@
    Copyright (C) James Peach 2006
    
    This library is free software; you can redistribute it and/or
-   modify it under the terms of the GNU Library General Public
+   modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
-   version 2 of the License, or (at your option) any later version.
+   version 3 of the License, or (at your option) any later version.
    
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
    
-   You should have received a copy of the GNU Library General Public
-   License along with this library; if not, write to the
-   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-   Boston, MA  02111-1307, USA.   
+   You should have received a copy of the GNU Lesser General Public License
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "winbind_client.h"
@@ -197,7 +195,7 @@ winbind_callback(nsd_file_t **rqp, int fd)
 
 	if (status != NSS_STATUS_SUCCESS) {
 		/* free any extra data area in response structure */
-		free_response(&response);
+		winbindd_free_response(&response);
 		nsd_logprintf(NSD_LOG_MIN, 
 			"callback (winbind) returning not found, status = %d\n",
 			status);
@@ -229,7 +227,7 @@ winbind_callback(nsd_file_t **rqp, int fd)
 			return NSD_ERROR;
 		}
 		
-		free_response(&response);
+		winbindd_free_response(&response);
 		
 		nsd_logprintf(NSD_LOG_MIN, "    %s\n", result);
 		nsd_set_result(rq, NS_SUCCESS, result, rlen, DYNAMIC);
@@ -254,7 +252,7 @@ winbind_callback(nsd_file_t **rqp, int fd)
 	        if (rlen == 0 || result == NULL)
 	            return NSD_ERROR;
 	    
-	        free_response(&response);
+	        winbindd_free_response(&response);
 	    
 	        nsd_logprintf(NSD_LOG_MIN, "    %s\n", result);
 	        nsd_set_result(rq, NS_SUCCESS, result, rlen, DYNAMIC);
@@ -281,7 +279,7 @@ winbind_callback(nsd_file_t **rqp, int fd)
 	        if (rlen == 0 || result == NULL)
 	            return NSD_ERROR;
 	    
-	        free_response(&response);
+	        winbindd_free_response(&response);
 	    
 	        nsd_logprintf(NSD_LOG_MIN, "    %s\n", result);
 	        nsd_set_result(rq, NS_SUCCESS, result, rlen, DYNAMIC);
@@ -292,7 +290,7 @@ winbind_callback(nsd_file_t **rqp, int fd)
 	    case WINBINDD_SETPWENT:
 		nsd_logprintf(NSD_LOG_MIN,
 			"callback (winbind) SETGRENT | SETPWENT\n");
-		free_response(&response);
+		winbindd_free_response(&response);
 		return(do_list(1,rq));
 
 	    case WINBINDD_GETGRENT:
@@ -313,7 +311,7 @@ winbind_callback(nsd_file_t **rqp, int fd)
 	            gr = (struct winbindd_gr *)response.extra_data.data;
 	            if (! gr ) {
 	                nsd_logprintf(NSD_LOG_MIN, "     no extra_data\n");
-	                free_response(&response);
+	                winbindd_free_response(&response);
 	                return NSD_ERROR;
 	            }
 	    
@@ -340,7 +338,7 @@ winbind_callback(nsd_file_t **rqp, int fd)
 	        }
 	    
 	        entries = response.data.num_entries;
-	        free_response(&response);
+	        winbindd_free_response(&response);
 	        if (entries < MAX_GETPWENT_USERS)
 	            return(do_list(2,rq));
 	        else
@@ -362,7 +360,7 @@ winbind_callback(nsd_file_t **rqp, int fd)
 		    pw = (struct winbindd_pw *)response.extra_data.data;
 		    if (! pw ) {
 			nsd_logprintf(NSD_LOG_MIN, "     no extra_data\n");
-			free_response(&response);
+			winbindd_free_response(&response);
 			return NSD_ERROR;
 		    }
 		    for (i = 0; i < response.data.num_entries; i++) {
@@ -387,7 +385,7 @@ winbind_callback(nsd_file_t **rqp, int fd)
 		}
 
 		entries = response.data.num_entries;
-		free_response(&response);
+		winbindd_free_response(&response);
 		if (entries < MAX_GETPWENT_USERS)
 		    return(do_list(2,rq));
 		else
@@ -398,11 +396,11 @@ winbind_callback(nsd_file_t **rqp, int fd)
 	    case WINBINDD_ENDPWENT:
 		nsd_logprintf(NSD_LOG_MIN, "callback (winbind) ENDGRENT | ENDPWENT\n");
 		nsd_append_element(rq, NS_SUCCESS, "\n", 1);
-		free_response(&response);
+		winbindd_free_response(&response);
 		return NSD_NEXT;
 
 	    default:
-		free_response(&response);
+		winbindd_free_response(&response);
 		nsd_logprintf(NSD_LOG_MIN, "callback (winbind) invalid command %d\n", (int)rq->f_cmd_data);
 		return NSD_NEXT;
 	}
