@@ -5,7 +5,7 @@
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
@@ -14,16 +14,16 @@
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "includes.h"
 
 /****************************************************************************
-normalise for DOS usage 
+ Normalise for DOS usage.
 ****************************************************************************/
-static void disk_norm(BOOL small_query, SMB_BIG_UINT *bsize,SMB_BIG_UINT *dfree,SMB_BIG_UINT *dsize)
+
+static void disk_norm(bool small_query, SMB_BIG_UINT *bsize,SMB_BIG_UINT *dfree,SMB_BIG_UINT *dsize)
 {
 	/* check if the disk is beyond the max disk size */
 	SMB_BIG_UINT maxdisksize = lp_maxdisksize();
@@ -36,11 +36,11 @@ static void disk_norm(BOOL small_query, SMB_BIG_UINT *bsize,SMB_BIG_UINT *dfree,
 		   errors */
 	}  
 
-	while (*dfree > WORDMAX || *dsize > WORDMAX || *bsize < 512) {
-		*dfree /= 2;
-		*dsize /= 2;
-		*bsize *= 2;
-		if(small_query) {	
+	if(small_query) {	
+		while (*dfree > WORDMAX || *dsize > WORDMAX || *bsize < 512) {
+			*dfree /= 2;
+			*dsize /= 2;
+			*bsize *= 2;
 			/*
 			 * Force max to fit in 16 bit fields.
 			 */
@@ -59,17 +59,17 @@ static void disk_norm(BOOL small_query, SMB_BIG_UINT *bsize,SMB_BIG_UINT *dfree,
 
 
 /****************************************************************************
-  return number of 1K blocks available on a path and total number 
+ Return number of 1K blocks available on a path and total number.
 ****************************************************************************/
 
-static SMB_BIG_UINT disk_free(const char *path, BOOL small_query, 
+SMB_BIG_UINT sys_disk_free(connection_struct *conn, const char *path, bool small_query, 
                               SMB_BIG_UINT *bsize,SMB_BIG_UINT *dfree,SMB_BIG_UINT *dsize)
 {
-	int dfree_retval;
+	SMB_BIG_UINT dfree_retval;
 	SMB_BIG_UINT dfree_q = 0;
 	SMB_BIG_UINT bsize_q = 0;
 	SMB_BIG_UINT dsize_q = 0;
-	char *dfree_command;
+	const char *dfree_command;
 
 	(*dfree) = (*dsize) = 0;
 	(*bsize) = 512;
@@ -78,13 +78,21 @@ static SMB_BIG_UINT disk_free(const char *path, BOOL small_query,
 	 * If external disk calculation specified, use it.
 	 */
 
-	dfree_command = lp_dfree_command();
+	dfree_command = lp_dfree_command(SNUM(conn));
 	if (dfree_command && *dfree_command) {
 		const char *p;
-		char **lines;
-		pstring syscmd;
+		char **lines = NULL;
+		char *syscmd = NULL;
+
+		syscmd = talloc_asprintf(talloc_tos(),
+				"%s %s",
+				dfree_command,
+				path);
+
+		if (!syscmd) {
+			return (SMB_BIG_UINT)-1;
+		}
 
-		slprintf(syscmd, sizeof(syscmd)-1, "%s %s", dfree_command, path);
 		DEBUG (3, ("disk_free: Running command %s\n", syscmd));
 
 		lines = file_lines_pload(syscmd, NULL);
@@ -115,10 +123,19 @@ static SMB_BIG_UINT disk_free(const char *path, BOOL small_query,
 		} else {
 			DEBUG (0, ("disk_free: sys_popen() failed for command %s. Error was : %s\n",
 				syscmd, strerror(errno) ));
-			sys_fsusage(path, dfree, dsize);
+			if (sys_fsusage(path, dfree, dsize) != 0) {
+				DEBUG (0, ("disk_free: sys_fsusage() failed. Error was : %s\n",
+					strerror(errno) ));
+				return (SMB_BIG_UINT)-1;
+			}
 		}
-	} else
-		sys_fsusage(path, dfree, dsize);
+	} else {
+		if (sys_fsusage(path, dfree, dsize) != 0) {
+			DEBUG (0, ("disk_free: sys_fsusage() failed. Error was : %s\n",
+				strerror(errno) ));
+			return (SMB_BIG_UINT)-1;
+		}
+	}
 
 	if (disk_quotas(path, &bsize_q, &dfree_q, &dsize_q)) {
 		(*bsize) = bsize_q;
@@ -133,10 +150,10 @@ static SMB_BIG_UINT disk_free(const char *path, BOOL small_query,
 	}
 
 	if ((*dsize)<1) {
-		static int done;
+		static bool done = false;
 		if (!done) {
 			DEBUG(0,("WARNING: dfree is broken on this system\n"));
-			done=1;
+			done=true;
 		}
 		*dsize = 20*1024*1024/(*bsize);
 		*dfree = MAX(1,*dfree);
@@ -153,12 +170,54 @@ static SMB_BIG_UINT disk_free(const char *path, BOOL small_query,
 	return(dfree_retval);
 }
 
-
 /****************************************************************************
-wrap it to get filenames right
+ Potentially returned cached dfree info.
 ****************************************************************************/
-SMB_BIG_UINT sys_disk_free(const char *path, BOOL small_query, 
-                           SMB_BIG_UINT *bsize,SMB_BIG_UINT *dfree,SMB_BIG_UINT *dsize)
+
+SMB_BIG_UINT get_dfree_info(connection_struct *conn,
+			const char *path,
+			bool small_query,
+			SMB_BIG_UINT *bsize,
+			SMB_BIG_UINT *dfree,
+			SMB_BIG_UINT *dsize)
 {
-	return disk_free(path,small_query, bsize,dfree,dsize);
+	int dfree_cache_time = lp_dfree_cache_time(SNUM(conn));
+	struct dfree_cached_info *dfc = conn->dfree_info;
+	SMB_BIG_UINT dfree_ret;
+
+	if (!dfree_cache_time) {
+		return SMB_VFS_DISK_FREE(conn,path,small_query,bsize,dfree,dsize);
+	}
+
+	if (dfc && (conn->lastused - dfc->last_dfree_time < dfree_cache_time)) {
+		/* Return cached info. */
+		*bsize = dfc->bsize;
+		*dfree = dfc->dfree;
+		*dsize = dfc->dsize;
+		return dfc->dfree_ret;
+	}
+
+	dfree_ret = SMB_VFS_DISK_FREE(conn,path,small_query,bsize,dfree,dsize);
+
+	if (dfree_ret == (SMB_BIG_UINT)-1) {
+		/* Don't cache bad data. */
+		return dfree_ret;
+	}
+
+	/* No cached info or time to refresh. */
+	if (!dfc) {
+		dfc = TALLOC_P(conn->mem_ctx, struct dfree_cached_info);
+		if (!dfc) {
+			return dfree_ret;
+		}
+		conn->dfree_info = dfc;
+	}
+
+	dfc->bsize = *bsize;
+	dfc->dfree = *dfree;
+	dfc->dsize = *dsize;
+	dfc->dfree_ret = dfree_ret;
+	dfc->last_dfree_time = conn->lastused;
+
+	return dfree_ret;
 }
