@@ -5,7 +5,7 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
@@ -14,8 +14,7 @@
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
 /* 
  * A function of this type is passed to the '
@@ -23,16 +22,49 @@
  * include
  */
 
-typedef NTSTATUS (*rpc_command_fn)(const DOM_SID *, const char *, 
-				   struct cli_state *, TALLOC_CTX *, int, const char **);
-				   
+#include "lib/netapi/netapi.h"
+#include "libnet/libnet.h"
+
+typedef NTSTATUS (*rpc_command_fn)(const DOM_SID *,
+				const char *, 
+				struct cli_state *cli,
+				struct rpc_pipe_client *,
+				TALLOC_CTX *,
+				int,
+				const char **);
+
 typedef struct copy_clistate {
 	TALLOC_CTX *mem_ctx;
 	struct cli_state *cli_share_src;
 	struct cli_state *cli_share_dst;
 	char *cwd;
 	uint16 attribute;
-}copy_clistate; 
+}copy_clistate;
+
+struct rpc_sh_ctx {
+	struct cli_state *cli;
+
+	DOM_SID *domain_sid;
+	const char *domain_name;
+
+	const char *whoami;
+	const char *thiscmd;
+	struct rpc_sh_cmd *cmds;
+	struct rpc_sh_ctx *parent;
+};
+
+struct rpc_sh_cmd {
+	const char *name;
+	struct rpc_sh_cmd *(*sub)(TALLOC_CTX *mem_ctx,
+				  struct rpc_sh_ctx *ctx);
+	int pipe_idx;
+	NTSTATUS (*fn)(TALLOC_CTX *mem_ctx, struct rpc_sh_ctx *ctx,
+		       struct rpc_pipe_client *pipe_hnd,
+		       int argc, const char **argv);
+	const char *help;
+};
+
+enum netdom_domain_t { ND_TYPE_NT4, ND_TYPE_AD };
 
 /* INCLUDE FILES */
 
@@ -40,24 +72,15 @@ typedef struct copy_clistate {
  
 /* MACROS & DEFINES */
 
-#define NET_FLAGS_MASTER 1
-#define NET_FLAGS_DMB 2
-
-/* Would it be insane to set 'localhost' as the default
-   remote host for this operation? 
-
-   For example, localhost is insane for a 'join' operation.
-*/
-#define NET_FLAGS_LOCALHOST_DEFAULT_INSANE 4 
-
-/* We want to find the PDC only */
-#define NET_FLAGS_PDC 8 
-
-/* We want an anonymous connection */
-#define NET_FLAGS_ANONYMOUS 16 
-
-/* don't open an RPC pipe */
-#define NET_FLAGS_NO_PIPE 32
+#define NET_FLAGS_MASTER 			0x00000001
+#define NET_FLAGS_DMB 				0x00000002
+#define NET_FLAGS_LOCALHOST_DEFAULT_INSANE 	0x00000004 	/* Would it be insane to set 'localhost'
+								   as the default remote host for this
+								   operation?  For example, localhost
+								   is insane for a 'join' operation.  */
+#define NET_FLAGS_PDC 				0x00000008	/* PDC only */ 
+#define NET_FLAGS_ANONYMOUS 			0x00000010	/* use an anonymous connection */
+#define NET_FLAGS_NO_PIPE 			0x00000020	/* don't open an RPC pipe */
 
 /* net share operation modes */
 #define NET_MODE_SHARE_MIGRATE 1
@@ -80,10 +103,10 @@ extern int opt_timeout;
 extern const char *opt_host;
 extern const char *opt_user_name;
 extern const char *opt_password;
-extern BOOL opt_user_specified;
+extern bool opt_user_specified;
 
-extern BOOL opt_localgroup;
-extern BOOL opt_domaingroup;
+extern int opt_localgroup;
+extern int opt_domaingroup;
 extern const char *opt_newntname;
 extern int opt_rid;
 extern int opt_acls;
@@ -91,9 +114,11 @@ extern int opt_attrs;
 extern int opt_timestamps;
 extern const char *opt_exclude;
 extern const char *opt_destination;
+extern int opt_testmode;
 
-extern BOOL opt_have_ip;
-extern struct in_addr opt_dest_ip;
+extern int opt_have_ip;
+extern struct sockaddr_storage opt_dest_ip;
+extern struct libnetapi_ctx *netapi_ctx;
 
 extern const char *share_type[];
 
